--[[

	TradesBar for WoW (World of Warcraft)
	Desc:	A row of buttons for your Trades Skills

	LIST OF AUTHORS and FIXERS
	Author 1:	Ian 'CRiSPy' Floyd	"Iling on Skywall(Nightelf)"
	Author 2:	David Carr			"Groz on Shattered Hand(Orc)"
	Author 3:	Eric Gingell			"Treader (plus many others) on Cenarion Circle (Night Elf)"
	German Tanslator1: Farook of wowinterface.com forum
	French Tanslator2: Trucifix of ui.worldofwar.net
	Chinese Translator: norova of wowinterface.com
	Simplified Chinese localized by Yeachan @ CWDG
	To do:
		Use OnTooltipSetSpell with tooltip:GetSpell() to add skill levels (maybe)
--]]

local aName = "TradesBar"
local aVer = GetAddOnMetadata(aName, "Version")
local TradesLocale = GetLocale()
local player = UnitName('player')
local _G = _G
local tradesCount
local SpellInfo = {}
local TRADESBAR_ADMIN_MAX_BUTTONS = 20

function TradesBar_GetSpellInfo(spellID)
	local spellName = GetSpellInfo(spellID)
	if spellName then
		SpellInfo[spellName] = spellID
	end
	return spellName;
end

if not GetSpellName then
	GetSpellName = GetSpellBookItemName
end

local TRADES_ALCHEMY = TradesBar_GetSpellInfo(2259)
local TRADES_BLACKSMITHING = TradesBar_GetSpellInfo(3100)
local TRADES_COOKING = TradesBar_GetSpellInfo(2550)
local TRADES_BASIC_CAMPFIRE = TradesBar_GetSpellInfo(818)
local TRADES_ENCHANTING = TradesBar_GetSpellInfo(7411)
local TRADES_DISENCHANT = TradesBar_GetSpellInfo(13262)
local TRADES_ENGINEERING = TradesBar_GetSpellInfo(4036)
local TRADES_FIRST_AID = TradesBar_GetSpellInfo(7162)
local TRADES_FISHING = TradesBar_GetSpellInfo(7620)
local TRADES_HERBALISM = TradesBar_GetSpellInfo(9134)
local TRADES_FIND_HERBS = TradesBar_GetSpellInfo(2383)
local TRADES_FIND_TREASURE = TradesBar_GetSpellInfo(2481)
local TRADES_LEATHERWORKING = TradesBar_GetSpellInfo(3104)
local TRADES_PICK_LOCK = TradesBar_GetSpellInfo(1804)
local TRADES_PICK_POCKET = TradesBar_GetSpellInfo(921)
local TRADES_FIND_MINERALS = TradesBar_GetSpellInfo(2580)
local TRADES_SMELTING = TradesBar_GetSpellInfo(2656)
local TRADES_MINING = TradesBar_GetSpellInfo(2575)
local TRADES_SKINNING = TradesBar_GetSpellInfo(8613)
local TRADES_TAILORING = TradesBar_GetSpellInfo(3908)
local TRADES_POISONS = TradesBar_GetSpellInfo(2842)
local TRADES_JEWELCRAFTING = TradesBar_GetSpellInfo(25229)
local TRADES_PROSPECTING = TradesBar_GetSpellInfo(31252)
local TRADES_BT = TradesBar_GetSpellInfo(5149)
local TRADES_DISARM = TradesBar_GetSpellInfo(1842)
local TRADES_FEEDPET = TradesBar_GetSpellInfo(6991)
local TRADES_SHADOWEAVE_TAILORING = TradesBar_GetSpellInfo(26800)
local TRADES_GNOME_ENGINEERING = TradesBar_GetSpellInfo(20220)
local TRADES_GOBLIN_ENGINEERING = TradesBar_GetSpellInfo(20221)
local TRADES_DRAGONSCALE_LEATHERWORKING = TradesBar_GetSpellInfo(10656)
local TRADES_ELEMENTAL_LEATHERWORKING = TradesBar_GetSpellInfo(10658)
local TRADES_TRIBAL_LEATHERWORKING = TradesBar_GetSpellInfo(10660)
local TRADES_SPELLFIRE_TAILORING = TradesBar_GetSpellInfo(26796)
local TRADES_HEARTHSTONE = TradesBar_GetSpellInfo(8690)
local TRADES_INSCRIPTION = TradesBar_GetSpellInfo(45358)
local TRADES_MILLING = TradesBar_GetSpellInfo(51005)
local TRADES_CALLPET = TradesBar_GetSpellInfo(883)
local TRADES_DISMISSPET = TradesBar_GetSpellInfo(2641)
local TRADES_RUNEFORGING = TradesBar_GetSpellInfo(53424)
local TRADES_DEATHGATE = TradesBar_GetSpellInfo(50977)
local TRADES_ARECALL = TradesBar_GetSpellInfo(556)
local TRADES_LOCKPICKING = GetSpellInfo(1809)

BINDING_HEADER_TRADES = "TradesBar";

BINDING_NAME_TRADES_ALCHEMY = TRADES_ALCHEMY
BINDING_NAME_TRADES_BLACKSMITHING = TRADES_BLACKSMITHING
BINDING_NAME_TRADES_COOKING = TRADES_COOKING
BINDING_NAME_TRADES_ENCHANTING = TRADES_ENCHANTING
BINDING_NAME_TRADES_ENGINEERING = TRADES_ENGINEERING
BINDING_NAME_TRADES_FIRST_AID = TRADES_FIRST_AID
BINDING_NAME_TRADES_LEATHERWORKING = TRADES_LEATHERWORKING
BINDING_NAME_TRADES_SMELTING = TRADES_SMELTING
BINDING_NAME_TRADES_TAILORING = TRADES_TAILORING
BINDING_NAME_TRADES_POISONS = TRADES_POISONS
BINDING_NAME_TRADES_JEWELCRAFTING = TRADES_JEWELCRAFTING

BINDING_NAME_TRADES_INSCRIPTION = TRADES_INSCRIPTION

-- Populate _G with locale data
Trades_LoadLocale(TradesLocale)

--[[
local Trades = {
	{
		title = (.*),		-- The name of the skill or item. Use "item:{itemID}" and "GetSpellInfo({spellID})" whenever possible
		panel = [45]{1},	-- The panel the admin checkbox will be attached to on the admin frame
		[type = (.*),]		-- The type of button this will be (item, spell, or skill). Default: "spell"
	},
}
--]]

TradesBarData = { };
-- panel = 6 makes it so the button won't show on the config frame.
local Trades = {
	{
		title = TRADES_ALCHEMY,
		panel = 4,
--		type = 'skill',
	},
	{
		title = TRADES_BLACKSMITHING,
		panel = 4,
--		type = 'skill',
	},
	{
		title = TRADES_COOKING,
		panel = 4,
--		type = 'skill',
	},
	{
		title = TRADES_ENCHANTING,
--		type = 'skill',
		panel = 4,
	},
	{
		title = TRADES_BASIC_CAMPFIRE,
		panel = 4,
	},
	{
		title = TRADES_DISENCHANT,
		panel = 4,
	},
	{
		title = TRADES_ENGINEERING,
--		type = 'skill',
		panel = 4,
	},
	{
		title = TRADES_FIRST_AID,
--		type = 'skill',
		panel = 4,
	},
	{
		title = TRADES_FISHING,
		panel = 4,
	},
	{
		title = TRADES_HERBALISM,
		panel = 4,
	},
	{
		title = TRADES_FIND_HERBS,
		panel = 4,
	},
	{
		title = TRADES_FIND_TREASURE,
		panel = 4,
	},
	{
		title = TRADES_LEATHERWORKING,
--		type = 'skill',
		panel = 4,
	},
	{
		title = TRADES_PICK_POCKET,
		panel = 4,
	},
	{
		title = TRADES_PICK_LOCK,
		panel = 4,
	},
	{
		title = TRADES_FIND_MINERALS,
		panel = 4,
	},
	{
		title = TRADES_SMELTING,
--		type = 'skill',
		panel = 4,
	},
	{
		title = TRADES_SKINNING,
		panel = 4,
	},
	{
		title = TRADES_TAILORING,
--		type = 'skill',
		panel = 4,
	},
	{
		title = TRADES_POISONS,
--		type = 'skill',
		panel = 4,
	},
	{
		title = TRADES_JEWELCRAFTING,
--		type = 'skill',
		panel = 4,
	},
	{
		title = TRADES_BT,
--		type = 'skill',
		panel = 4,
	},
	{
		title = TRADES_MINING,
		panel = 4,
	},
	{
		title = TRADES_PROSPECTING,
		panel = 4,
	},
	{
		title = TRADES_DISARM,
		panel = 4,
	},
	{
		title = TRADES_HEARTHSTONE,
		type = 'item',
		panel = 4,
	},
	{
		title = TRADES_FEEDPET,
		panel = 4,
	},
	{
		title = TRADES_CALLPET,
		panel = 4,
	},
	{
		title = TRADES_DISMISSPET,
		panel = 4,
	},
	{
		title = TRADES_RUNEFORGING;
--		type = 'skill',
		panel = 4,
	},
	{
		title = TRADES_DEATHGATE;
		panel = 4,
	},
	{
		title = TRADES_ARECALL;
		panel = 4,
	},
	{
		title = TRADES_INSCRIPTION,
		panel = 4,
--		type = "skill",
	},
	{
		title = TRADES_MILLING,
		panel = 4,
	},
	-- Specialty skills are basically the same as the original from a Lua point of view, but they have a different
	-- texture and a diffent tooltip.
	{
		title = TRADES_SHADOWEAVE_TAILORING;
--		type = 'skill',
		panel = 6,
	},
	{
		title = TRADES_DRAGONSCALE_LEATHERWORKING;
--		type = 'skill',
		panel = 6,
	},
	{
		title = TRADES_ELEMENTAL_LEATHERWORKING;
--		type = 'skill',
		panel = 6,
	},
	{
		title = TRADES_SPELLFIRE_TAILORING;
--		type = 'skill',
		panel = 6,
	},
	{
		title = TRADES_TRIBAL_LEATHERWORKING;
--		type = 'skill',
		panel = 6,
	},
	{
		title = TRADES_GNOME_ENGINEERING;
--		type = 'skill',
		panel = 6,
	},
	{
		title = TRADES_GOBLIN_ENGINEERING;
--		type = 'skill',
		panel = 6,
	},
	{
		title = "Random Mount";
		type = "special";
		subtype = "MOUNT";
		panel = 4,
	},
	{
		title = "Random Pet";
		type = "special";
		subtype = "CRITTER";
		panel = 4,
	},
};

local TradesBar_Buttons = {};
local AdminPanels = {};
local original_heights = {}
TradesBar_Debug = false

local TRADESBAR_ENABLED = true;
local TradesBar_UpdateInterval = 1.0;
local TradesBar_TooltipRanksText = {}

local color = {
	RED			= "ff0000"; -- current rank <= 25% of max
	ORANGE		= "ff7f00"; -- current rank <= 50% and > 25% of max
	YELLOW		= "ffff00"; -- current rank <= 75% and > 50% of max
	GREEN		= "00ff00"; -- current rank <= 100% and > 75% of max
}

local TradesBar_is_fishingbuddy = false;
local inCombat, playerName, vMounted, button;
playerName = player .. " of " .. GetCVar("realmName");

function TradesBar_Binding(f)
	-- ??
end

function TradesBar_AddSpellOrItem()
	if not CursorHasItem() and not CursorHasSpell() then
		return
	end
	local infoType, info1, info2, info3 = GetCursorInfo()
	ClearCursor()
	if infoType == 'item' then
		local itemName = GetItemInfo(info1)
		if itemName then
			TradesBar_AddItem(itemName)
		end
	elseif infoType == 'spell' then
		if info1 == 0 then -- As of 3.0.2 GetCursorInfo() returns "spell", 0, "spell" for mini-pets and mounts. Alert the user and return.
			return DEFAULT_CHAT_FRAME:AddMessage("Invalid spell ID, this is probably a mount or pet. These are added automatically.", 1.0, 0.0, 0.25);
		end
		local spellName = GetSpellName(info1, 'spell');
		TradesBar_AddSpell(spellName)
	end
end

function TradesBar_AddSpell(spell)
	local i = 1;
	while true do
		local spellName = GetSpellName(i, BOOKTYPE_SPELL);
		if (not spellName) then
			break;
		end
		if strfind(strlower(spellName), strlower(spell)) then
			if (TradesBar_Debug) then
				DEFAULT_CHAT_FRAME:AddMessage("Found |cffffffff'"..spellName.."'|r at slot |cffffffff'"..i.."'", 0.1, 1.0, 0.0);
			end
			local spellTable = {
				title = spellName,
				panel = 5,
			}
			if not TradesBar_InTable(Trades, spellName) then
				table.insert(Trades, spellTable); -- Add to "Trades" for real-time adding of skills.
				table.insert(TradesBarCustomSpells, spellTable);
				if not TradesBarData["skills"][spellName] then
					TradesBarData["skills"][spellName] = {}
				end
				TradesBarData["skills"][spellName].visible = "true"
				if not AdminPanels[5] then
					AdminPanels[5] = {};
				end
				table.insert(AdminPanels[5], spellName);
			end
		end
		i = i + 1;
		if (i > 250) then
			break;
		end
	end
	TradesBar_Load()
--	TradesBarConfig_OnShow(TradesBarConfig)
end

function TradesBar_AddItem(item)
	-- The item *must* be in your local cache for this to work.
	local itemName, itemLink = GetItemInfo(item)
	if itemLink then
		local itemWord, itemId
		local itemString = strmatch(itemLink, "^|c%x+|H(.+)|h%[.+%]")
		if itemString then
			itemId = strmatch(itemString, "item:(%d+):.+")
--			ChatFrame1:AddMessage('item:'..itemId)
			local itemTable = {
				title = 'item:'..itemId,
				type = 'item',
				panel = 5,
			}
			if TradesBar_InTable(Trades, 'item:'..itemId) == false then
				table.insert(Trades, itemTable); -- Add to "Trades" for real-time adding of items.
				table.insert(TradesBarCustomSpells, itemTable);
				if not TradesBarData["skills"]['item:'..itemId] then
					TradesBarData["skills"]['item:'..itemId] = {}
				end
				TradesBarData["skills"]['item:'..itemId].visible = "true"
				if not AdminPanels[5] then
					AdminPanels[5] = {};
				end
				table.insert(AdminPanels[5], itemTable);
			end
			TradesBar_Load()
--			TradesBarConfig_OnShow(TradesBarConfig)
		end
	end
end

function TradesBar_GetSpecialButton(ptype, buttonName)
	local spellID, icon
	local button = CreateFrame("Button", buttonName, TradesBar, "TradesBarButtonTemplate");
	button.newMountID = 1
	button.companions = GetNumCompanions(ptype)
	button:RegisterEvent('COMPANION_LEARNED');
	button:RegisterEvent('COMPANION_UNLEARNED');
	
	button:SetScript("OnLeave", function(self)
		self.myTooltip = nil
		GameTooltip:Hide()
	end)
	
	button:SetScript("OnEvent", function(self, ...)
		self.companions = GetNumCompanions(ptype)
	end)
	
	button:SetScript("OnEnter", function(self)
		self.myTooltip = 1
		GameTooltip_SetDefaultAnchor(GameTooltip, self)
		local colors = ChatTypeInfo["SYSTEM"]
		if not self.spinning then
			GameTooltip:SetHyperlink("spell:" .. spellID)
			GameTooltip:AddLine("Left-Click to start spinning without summoning or summon the current companion at the time of the click.", colors.r, colors.g, colors.b, colors.id)
			GameTooltip:AddLine(" ")
			GameTooltip:AddLine("Right-Click to stop spinning without summoning or bypass the spinning and summon the companion.", colors.r, colors.g, colors.b, colors.id)
			GameTooltip:AddLine(" ")
			GameTooltip:AddLine("== Your " .. (ptype == "MOUNT" and "Mounts" or "Pets") .. " ==")
			for i = 1, self.companions do
				local _, name = GetCompanionInfo(ptype, i)
				if name then
					GameTooltip:AddLine(name)
				end
			end
		else
			GameTooltip:AddLine("Spinning...", colors.r, colors.g, colors.b, colors.id)
		end
		GameTooltip:Show()
	end)
	
	button:SetScript("OnUpdate", function(self, elapsed)
		if self.spinning then
			self.newMountID = (self.newMountID % self.companions) + 1
			self.timeSince = 0
			if GameTooltip:IsShown() and self.myTooltip then
				self:GetScript("OnEnter")(self)
			end
		end
		if self.prevMountID ~= self.newMountID then
			self:ChangeMountID(self.newMountID)
			self.prevMountID = self.newMountID
		end
	end)
	button:SetScript("OnClick", function(self, button)
		if not IsShiftKeyDown() and not IsControlKeyDown() and not IsAltKeyDown() then
			-- Covers Left-Click while spinning and Right-Click while not spinning
			if (self.spinning and button == "LeftButton") or (not self.spinning and button == "RightButton") then
				self:ChangeMountID(self.newMountID)
				CallCompanion(ptype, self.newMountID)
				self.spinning = nil
				if GameTooltip:IsShown() and self.myTooltip then
					self:GetScript("OnEnter")(self)
				end
			-- Covers Right-Click while spinning.
			elseif button == "RightButton" then
				self.spinning = nil
				if GameTooltip:IsShown() and self.myTooltip then
					self:GetScript("OnEnter")(self)
				end
			-- Covers Left-Click while not spinning
			else
				self.spinning = 1
			end
		end
	end)
	
	function button:ChangeMountID(newMountID)
		spellID, icon = select(3, GetCompanionInfo(ptype, newMountID))
		if not InCombatLockdown() then
			self:SetID(newMountID)
			_G[self:GetName() .. "Icon"]:SetTexture(icon)
--		else
--			self.newMountID = (self.newMountID % GetNumCompanions(ptype)) + 1
--			self:ChangeMountID(self.newMountID)
		end
	end
	
	button:ChangeMountID(button.newMountID)
	return button
end

function TradesBar_MoveFrame(frame, point1, rel, point2, x, y)
	if frame and x ~= nil and y ~= nil and point1 ~= nil and point2 ~= nil and rel ~= nil then
		frame:ClearAllPoints()
		frame:SetPoint(point1, rel, point2, x, y)
	end
end

--TradesBarCustomSpells = {};
StaticPopupDialogs["TradesBar_AddSpell"] = {
	text = TRADES_TEXT_ADD_SPELL,
	button1 = OKAY,
	button2 = CANCEL,
	hasEditBox = 1,
	maxLetters = 50,
	OnAccept = function()
		local editBox = _G[this:GetParent():GetName().."EditBox"];
		TradesBar_AddSpell(editBox:GetText());
--		local dialog = StaticPopup_Show('TradesBar_ReloadNotice', editBox:GetText());
		if dialog then
			dialog.data = editBox:GetText();
		end
	end,
	OnShow = function()
		_G[this:GetName().."EditBox"]:SetFocus();
	end,
	OnHide = function()
		local ChatFrameEditBox = ChatEdit_FocusActiveWindow()
		if ( ChatFrameEditBox:IsVisible() ) then
			ChatEdit_FocusActiveWindow()
		end
		_G[this:GetName().."EditBox"]:SetText("");
	end,
	EditBoxOnEnterPressed = function()
		local editBox = _G[this:GetParent():GetName().."EditBox"];
		TradesBar_AddSpell(editBox:GetText());
		local dialog = StaticPopup_Show('TradesBar_ReloadNotice', editBox:GetText());
		if dialog then
			dialog.data = editBox:GetText();
		end
	end,
	EditBoxOnEscapePressed = function()
		this:GetParent():Hide();
	end,
	timeout = 0,
	exclusive = 1,
	whileDead = 1
};

StaticPopupDialogs["TradesBar_LoadDefault"] = {
	text = format(TRADES_TEXT_LOAD_DEFAULT, '"'..YES..'"'),
	button1 = YES,
	button2 = CANCEL,
	OnAccept = function()
		TradesBarData = nil;
		TradesBarCustomSpells = {};
		ReloadUI();
	end,
	OnCancel = function()
		this:GetParent():Hide();
	end,
	EditBoxOnEscapePressed = function()
		this:GetParent():Hide();
	end,
	timeout = 0,
	exclusive = 1,
	whileDead = 1
};

TradesBar_ATSW_IsEnabled_Hook = nil;

if (IsAddOnLoaded('FishingBuddy')) then
	TradesBar_is_fishingbuddy = true;
end

function TradesBar_OnLoad(this)
	-- Hooked for debugging
--	hooksecurefunc('SecureActionButton_OnClick',
--		function (self, button)
--			local spell = SecureButton_GetModifiedAttribute(self, "spell", button);
--			DEFAULT_CHAT_FRAME:AddMessage('Click accepted for button '..spell..'('..button..')');
--		end
--	);
	TradesBar_TimeSinceLastUpdate = 0;
	this:RegisterEvent("PLAYER_ENTERING_WORLD");
	this:RegisterEvent("VARIABLES_LOADED");
	this:RegisterEvent("SPELLS_CHANGED");
	this:RegisterEvent("PLAYER_AURAS_CHANGED");
	this:RegisterEvent("SPELL_UPDATE_COOLDOWN");
	this:RegisterEvent('ADDON_LOADED');
	this:RegisterEvent('COMPANION_UPDATE');
	
	SLASH_TRADESBAR1 = "/tradesbar";
	SLASH_TRADESBAR2 = "/tbar";
	SLASH_TRADESBAR3 = "/tb";
	SlashCmdList["TRADESBAR"] = function(msg)
		TradesBar_SlashHandler(string.lower(msg));
	end
end

function TradesBar_HelpMsg()
	for i = 1, 3 do
		DEFAULT_CHAT_FRAME:AddMessage(_G["TRADES_HELP_MESSAGE"..i]);
	end
end

function TradesBar_SlashHandler(msg)
	if (string.find(msg, "help")) then
		TradesBar_HelpMsg();
	elseif (string.find(msg, "toggle")) then
		TradesBar_Hide();
	elseif (string.find(msg, "default")) then
		StaticPopup_Show('TradesBar_LoadDefault');
	elseif (string.find(msg, "debug")) then
		TradesBarDump()
	elseif (string.find(msg, "admin") or string.find(msg, "config")) then
		if(	TradesBarConfig:IsVisible() ) then
			TradesBarConfig:Hide();
		else
			TradesBarConfig:Show();
		end
	else
		TradesBar_HelpMsg();
	end
--	tinsert(UISpecialFrames, "TradesBarConfig"); 
end

function StartTradesBar()
	if TradesBar_Debug then
		TradesBarConfig:Show();
	end
	TradesBar_UpdateCooldowns();
	if CustomTradesOrder then
		Trades = CustomTradesOrder;
	end
	for i in pairs(Trades) do
		if not TradesBarData["skills"][Trades[i].title] then
			TradesBarData["skills"][Trades[i].title] = {}
		end
		if (not TradesBarData["skills"][Trades[i].title].visible) then
			TradesBarData["skills"][Trades[i].title].visible = "true";
		end
	end
	TradesBar_Update();
	TradesBar_Scale();
end

function TradesBar_InTable(haystack, needle)
	for i in ipairs(haystack) do
		if haystack[i].title == needle then
			return i;
		end
	end
	return false;
end

do
	local mountTypes = {"MOUNT", "CRITTER"}
	function TradesBar_ScanPets()
		-- Mounts and Pets
		local creatureID, creatureName, spellID, icon, active
		for _, ctype in pairs(mountTypes) do
			for index = 1, GetNumCompanions(ctype) do
				creatureID, creatureName, spellID, icon, active = GetCompanionInfo(ctype, index)
				if creatureName and TradesBar_InTable(Trades, creatureName) == false then
					table.insert(Trades, {
						title = creatureName,
						type = "macro",
						macrotext = "/run CallCompanion('" .. ctype .. "'," .. index .. ")",
						texture = icon,
						panel = 4,
						mountOrPet = ctype,
						spellID = spellID,
					})
					if not TradesBarData["skills"][creatureName] then
						TradesBarData["skills"][creatureName] = {}
					end
					if not TradesBarData["skills"][creatureName].visible then
						TradesBarData["skills"][creatureName].visible = "true"
						TradesBarData["skills"][creatureName].buttonType = ctype
					end
					if TradesBarData["settings"]["hide_all_"..ctype] == "true" then
						TradesBarData["skills"][creatureName].visible = "false"
					end
				end
			end
		end
	end
end

local eventsFired = {}
function TradesBar_OnEvent(self, event, ...)
	eventsFired[event] = true
	if (event == "PLAYER_ENTERING_WORLD") then
		self:UnregisterEvent("PLAYER_ENTERING_WORLD");
		if CustomTradesBar and type(CustomTradesBar) == 'function' then
			CustomTradesBar();
		end
		-- Add the custom buttons
		if TradesBarCustomSpells then
			for i in ipairs(TradesBarCustomSpells) do
				if TradesBar_InTable(Trades, TradesBarCustomSpells[i].title) == false then
					if TradesBarCustomSpells[i].name then
						TradesBarCustomSpells[i].title = TradesBarCustomSpells[i].name
						TradesBarCustomSpells[i].name = nil
					end
					if (TradesBarCustomSpells[i].type == 'item' and GetItemInfo(TradesBarCustomSpells[i].title)) or TradesBarCustomSpells[i].type ~= 'item' then
						table.insert(Trades,
							{
								title = TradesBarCustomSpells[i].title,
								panel = 5,
								type = TradesBarCustomSpells[i].type,
							}
						);
					end
				else
					TradesBarCustomSpells[i] = nil;
				end
			end
		end
		-- Add the custom buttons
		if CustomTradesBarFromFile then
			for i in ipairs(CustomTradesBarFromFile) do
				if TradesBar_InTable(Trades, CustomTradesBarFromFile[i].title) == false then
					if (CustomTradesBarFromFile[i].type == 'item' and GetItemInfo(CustomTradesBarFromFile[i].title)) or CustomTradesBarFromFile[i].type ~= 'item' then
						table.insert(Trades,
							{
								title = CustomTradesBarFromFile[i].title,
								panel = 5,
								type = CustomTradesBarFromFile[i].type,
								texture = CustomTradesBarFromFile[i].texture,
								macrotext = CustomTradesBarFromFile[i].macrotext,
								tooltiptext = CustomTradesBarFromFile[i].tooltiptext,
							}
						);
					end
				else
					CustomTradesBarFromFile[i] = nil;
				end
			end
		end
		TradesBar_Load()
		StartTradesBar();
	elseif (event == "COMPANION_UPDATE") then
		TradesBar_ScanPets()
		TradesBar_Load()
		TradesBarConfig_OnShow(TradesBarConfig)
	elseif (event == "SPELLS_CHANGED" or event == "PLAYER_AURAS_CHANGED") then
		if eventsFired["PLAYER_ENTERING_WORLD"] then
			TradesBar_Update();
		end
	elseif (event == "SPELL_UPDATE_COOLDOWN") then
		TradesBar_UpdateCooldowns();
	-- If the following addons load after this one, set the variables here.
	elseif (event == 'ADDON_LOADED') then
		if arg1 == 'FishingBuddy' then
			TradesBar_is_fishingbuddy = true;
		end
	elseif ( event == 'VARIABLES_LOADED') then
		if not TradesBarCustomSpells then
			TradesBarCustomSpells = {}
		end
		if not TradesBar_Debug then
			TradesBar_Debug = false
		end
		local pname = player
		if (pname ~= UNKNOWNBEING and pname ~= UNKNOWNOBJECT and pname and not playerName) then
			playerName = pname.." of "..GetCVar("realmName");
		end
		if not TradesBarData then
			TradesBarData = {}
		end
		if not TradesBarData["skills"] then -- seems a little unnecessary
			TradesBarData["skills"] = {}
		end
		if not TradesBarData["settings"] then
			TradesBarData["settings"] = {
				["hidebar"] = "false",
				["horizontal"] = "true",
				["order"] = "az",
				["scale"] = 1,
			}
		end
		if TradesBarData[playerName] then
			TradesBarData["settings"] = {
				["hidebar"] = TradesBarData[playerName]["hidebar"],
				["horizontal"] = TradesBarData[playerName]["horizontal"],
				["order"] = TradesBarData[playerName]["order"],
				["scale"] = TradesBarData[playerName]["scale"],
			}
			TradesBarData["skills"] = {}
			for k in pairs(TradesBarData[playerName]) do
				if k ~= "version" and k ~= "hide_all_MOUNT" and k ~= "hide_all_CRITTER" and k ~= "hidebar" and k ~= "horizontal"  and k ~= "order"  and k ~= "scale" and k ~= "updated" and k ~= "left" and k ~= "top" then
					TradesBarData["skills"][k] = TradesBarData[playerName][k]
				end
				TradesBarData[playerName][k] = nil
			end
			TradesBarData[playerName] = nil
		end
		for k, v in pairs(TradesBarData["skills"]) do
			if TradesBarData["skills"][k] == "true" or TradesBarData["skills"][k] == "false" then
				TradesBarData["skills"][k] = {
					["visible"] = v,
					["key"] = nil,
				}
			end
		end
		TradesBar_MoveFrame(TradesBarConfigPanel1AddSpellButtonText1, 'TOPLEFT', TradesBarConfigPanel1AddSpellButton, 'TOPLEFT', -1 * TradesBarConfigPanel1AddSpellButtonText1:GetWidth() - 10, 2)
		TradesBar_MoveFrame(TradesBarConfigPanel1AddSpellButtonText2, 'TOPLEFT', TradesBarConfigPanel1AddSpellButton, 'TOPLEFT', -1 * TradesBarConfigPanel1AddSpellButtonText2:GetWidth() - 10, -10)
		TradesBarBindingFrame.playername = playerName
		TradesBar_ScanPets()
	end
end

function TradesBar_ImplodeList(char, ...)
	local ret = select(1, ...)
	for i = 2, select("#", ...) do
		ret = ret .. char .. " " .. select(i, ...)
	end
	return ret
end

local queueframe = CreateFrame("Frame")
function queueframe:Load()
	self.queueitems = {
		-- this is the proper order
		TradesBar_Load = false, -- executes secure code -- cannot be done in combat
		TradesBar_Update = false, -- executes secure code -- cannot be done in combat
		TradesBar_O = false, -- executes secure code -- cannot be done in combat
		TradesBar_Hide = false, -- shows/hides a secure frame -- cannot be done in combat
	}
	self:SetScript("OnUpdate", function(self, elapsed)
		if not InCombatLockdown() then
			for func, bool in pairs(self.queueitems) do
				if bool == true then
					_G[func]()
					self.queueitems[func] = false
				end
			end
			self:Hide()
		end
	end)
end
queueframe:Load()

do
	local OpenBagItems = {
		[TRADES_DISENCHANT] = true;
		[TRADES_FEEDPET] = true;
		[TRADES_PICK_LOCK] = true;
		[TRADES_PROSPECTING] = true;
		[TRADES_MILLING] = true;
	}
		
	function TradesBar_Load()
		if InCombatLockdown() then
			queueframe.queueitems.TradesBar_Load = true
			queueframe:Show()
			return
		end
	--	DEFAULT_CHAT_FRAME:AddMessage('Making buttons');
		-- Make the buttons
		tradesCount = #Trades
		for id in ipairs(Trades) do
			if not TradesBar_Buttons[id] then
				if Trades[id].type == "special" then
					TradesBar_Buttons[id] = TradesBar_GetSpecialButton(Trades[id].subtype, 'TradesBarButton'..id)
				else
					TradesBar_Buttons[id] = CreateFrame('Button', 'TradesBarButton'..id, TradesBar, 'TradesBarButtonTemplate');
				end
			end
			if not Trades[id].type then
				Trades[id].type = 'spell'
			end
			if SpellInfo[Trades[id].title] then
				TradesBar_Buttons[id].spellID = SpellInfo[Trades[id].title];
			else
				TradesBar_Buttons[id].spellID = Trades[id].spellID;
			end
			TradesBar_Buttons[id].type = Trades[id].type;
			TradesBar_Buttons[id].title = Trades[id].title;
			TradesBar_Buttons[id].mountOrPet = Trades[id].mountOrPet;
			TradesBar_Buttons[id].btnName = 'TradesBarButton'..id;
	--			ChatFrame1:AddMessage(Trades[id].title..': '..tostring(Trades[id].type))
			if Trades[id].type == 'item' then
				local _, itemLink = GetItemInfo(Trades[id].title)
	--				ChatFrame1:AddMessage(tostring(itemLink))
				TradesBar_Buttons[id].itemLink = itemLink;
			else
	--				TradesBar_Buttons[id].spellID = TradesBar_GetSpellID(Trades[id].title)
			end
			TradesBar_Buttons[id]:RegisterForClicks('AnyUp');
	--			TradesBar_Buttons[id]:SetScript('OnClick', TradesBar_Buttons[id]:GetScript('OnClick'));
			-- If it has a panel, then we need to tell the script to add it and where to add it to the config frame.
			if TradesBar_Debug then
				ChatFrame1:AddMessage('Adding AdminPanel['..tostring(Trades[id].panel)..']['..id..']')
			end
			if not Trades[id].adminCheck then
				Trades[id].adminCheck = 1
				if Trades[id].panel then
					if not AdminPanels[Trades[id].panel] then
						AdminPanels[Trades[id].panel] = {};
					end
					table.insert(AdminPanels[Trades[id].panel], {
						title = Trades[id].title;
						panel = Trades[id].panel;
						type = Trades[id].type;
						subtype = Trades[id].subtype;
					})
				else
					if not AdminPanels[5] then
						AdminPanels[5] = {};
					end
					table.insert(AdminPanels[5], {
						title = Trades[id].title;
						panel = Trades[id].panel;
						type = Trades[id].type;
					})
				end
			end
			if Trades[id].type ~= 'skill' and Trades[id].type ~= 'macro' and Trades[id].type ~= "special" then
				-- Required for SecureActionButtonTemplate
				TradesBar_Buttons[id]:SetAttribute("type", Trades[id].type or "spell");
				TradesBar_Buttons[id]:SetAttribute("shift-" .. (Trades[id].type or "spell") .. "2", ATTRIBUTE_NOOP);
				TradesBar_Buttons[id]:SetAttribute("ctrl-" .. (Trades[id].type or "spell") .. "2", ATTRIBUTE_NOOP);
				TradesBar_Buttons[id]:SetAttribute("alt-" .. (Trades[id].type or "spell") .. "2", ATTRIBUTE_NOOP);
				TradesBar_Buttons[id]:SetAttribute((Trades[id].type or "spell"), Trades[id].title);
			elseif Trades[id].type == 'macro' then
				TradesBar_Buttons[id]:SetAttribute("type", Trades[id].type);
				TradesBar_Buttons[id]:SetAttribute("shift-macro2", ATTRIBUTE_NOOP);
				TradesBar_Buttons[id]:SetAttribute("ctrl-macro2", ATTRIBUTE_NOOP);
				TradesBar_Buttons[id]:SetAttribute("alt-macro2", ATTRIBUTE_NOOP);
				TradesBar_Buttons[id].texture = Trades[id].texture
				TradesBar_Buttons[id].tooltiptext = Trades[id].tooltiptext
				TradesBar_Buttons[id].macrotext = Trades[id].macrotext
	--				TradesBar_Buttons[id]:SetTexture(Trades[id].texture)
			elseif Trades[id].type == "special" then -- beats wrapping it in another if block
			else
				TradesBar_Buttons[id]:SetScript("OnClick",
					function(self, ...)
						CastSpellByName(self.title) -- does not work for "actions"
					end
				)
			end
			-- This doesn't work, yet.
	--			BoundButtons['Bind '..Trades[id].title] = 'TradesBarButton'..Trades[id].title
			
			if OpenBagItems[Trades[id].title] == true then
				TradesBar_Buttons[id]:HookScript('OnMouseUp',
					function(self, button)
						if button == "RightButton" then
							return
						end
						if OneBagFrame then
							OneBagFrame:Show()
							return
						end
						if BaudBagFrame then
							local bag
							for i = 1, 6 do
								bag = _G["BBCont1_" .. i]
								if bag then
									bag:Show()
								end
							end
							return
						end
						OpenAllBags(true);
					end
				)
			end
			
	--		TradesBar_Buttons[id]:SetID(0);
	--[[
			local frame, x;
			if TradesBar_Buttons[id - 1] then
				-- Previous button?
				frame = TradesBar_Buttons[id - 1]:GetName();
				x = 10;
			else
				-- Relative to parent?
				frame = 'TradesBar';
				x = 1;
			end
			TradesBar_Buttons[id]:ClearAllPoints();
			TradesBar_Buttons[id]:SetPoint('LEFT', frame, 'RIGHT', x, 0);
	--			DEFAULT_CHAT_FRAME:AddMessage('Button'..TradesBar_Buttons[id]:GetName());
	--]]
		end
		if eventsFired["VARIABLES_LOADED"] then
			TradesBar_Update()
		end
	end
end

function TradesBar_Scale()
	TradesBar:SetScale(TradesBarData["settings"]["scale"] or 1);
end

-- Convienience functions
function TradesBar_HasSkill(skill)
	if GetSpellInfo(skill) then
		return true;
	else
		return false;
	end
end

function TradesBar_HasItem(button)
	if button.type ~= 'item' then
		return
	end
	local searchItem = GetItemInfo(button.title)
	for bag = 0, NUM_BAG_SLOTS do
		for slot = 1, GetContainerNumSlots(bag) do
			local itemlink = GetContainerItemLink(bag, slot)
			if itemlink then
				local item = GetItemInfo(itemlink)
				if searchItem == item then
					return true
				end
			end
		end
	end
end

function TradesBar_O()
	if InCombatLockdown() then
		queueframe.queueitems.TradesBar_O = true
		queueframe:Show()
		return
	end
	-- Default to horizontal
	local prev, button, point, relative, x, y, height, width;
	x, y = 0, 0
	height, width = 28, 28
	for i in ipairs(TradesBar_Buttons) do
		TradesBar_Buttons[i]:ClearAllPoints();
		-- Default to visible
		if not TradesBarData["skills"][Trades[i].title] then
			TradesBarData["skills"][Trades[i].title] = {}
		end
		if (not TradesBarData["skills"][Trades[i].title].visible) then
			TradesBarData["skills"][Trades[i].title].visible = "true"
		end
		TradesBar_SpecialCheck(Trades[i]);
		TradesBar_visableCheck(Trades[i]);
		if not (Trades[i].i_am_special or Trades[i].hide_button) and ((Trades[i].type and ((Trades[i].type == 'item' and TradesBar_HasItem(Trades[i])) or Trades[i].type == 'macro' or Trades[i].type == 'special')) or TradesBar_HasSkill(Trades[i].title)) then
			if (button) then
				--prev = button:GetName();
				prev = true;
			end
			button = TradesBar_Buttons[i];
			if button.type == 'macro' then
				button:SetAttribute('macro1', false)
				button:SetAttribute('macrotext', button.macrotext);
			end
			if not button.companions or button.companions > 0 then
--				DEFAULT_CHAT_FRAME:AddMessage('Button '..tostring(button.title));
--				DEFAULT_CHAT_FRAME:AddMessage('PrevButton '..prev);
				if (TradesBarData["settings"]["horizontal"] == "true" and TradesBarData["settings"]["order"] == "az") then
					point, relative = "TOPLEFT", "TOPLEFT";
					if prev then
						x = x + 28
						y = y + 0
						width = width + 28
					end
				elseif (TradesBarData["settings"]["horizontal"] == "true" and TradesBarData["settings"]["order"] == "za") then
					point, relative = "TOPRIGHT", "TOPRIGHT";
					if prev then
						x = x - 28
						y = y + 0
						width = width + 28
					end
				elseif (TradesBarData["settings"]["horizontal"] == "false" and TradesBarData["settings"]["order"] == "az") then
					point, relative = "TOPRIGHT", "TOPRIGHT";
					if prev then
						x = x + 0
						y = y - 28
						height = height + 28
					end
				elseif (TradesBarData["settings"]["horizontal"] == "false" and TradesBarData["settings"]["order"] == "za") then
					point, relative = "BOTTOMLEFT", "BOTTOMLEFT";
					if prev then
						x = x + 0
						y = y + 28
						height = height + 28
					end
				else
					point, relative = "TOPLEFT", "TOPLEFT";
					if prev then
						x = x + 28
						y = y + 0
						width = width + 28
					end
				end
				button:SetPoint(point, 'TradesBar', relative, x, y);
				TradesBar:SetWidth(width)
				TradesBar:SetHeight(height)
			end
		end
	end
end

function TradesBar_Hide()
	if InCombatLockdown() then
		queueframe.queueitems.TradesBar_Hide = true
		queueframe:Show()
		return
	end
	local frame = TradesBar;
	if (frame) then
		if( frame:IsVisible() ) then
			frame:Hide();
		else
			frame:Show();
		end
	end
end

function TradesBar_FishingBuddy_Over()
	local frame = FishingBuddyFrame;
	if (frame) then
		if( frame:IsVisible() ) then
			frame:Hide();
		else
			frame:Show();
		end
	end
end

function TradesBar_visableCheck(button)
	button.hide_button = true;
	local value = button.title
	-- Special conditions
	if (value == TRADES_GNOME_ENGINEERING or value == TRADES_GOBLIN_ENGINEERING) then
		value = TRADES_ENGINEERING;
	end
	if (value == TRADES_DRAGONSCALE_LEATHERWORKING or value == TRADES_ELEMENTAL_LEATHERWORKING or value == TRADES_TRIBAL_LEATHERWORKING) then
		value = TRADES_LEATHERWORKING;
	end
	if (value == TRADES_SHADOWEAVE_TAILORING or value == TRADES_SPELLFIRE_TAILORING) then
		value = TRADES_TAILORING;
	end
	if not TradesBarData["skills"][value] then
		TradesBarData["skills"][value] = {}
	end
	if (TradesBarData["skills"][value].visible == "true") then
		button.hide_button = false;
	end
end

function TradesBar_SpecialCheck(button)
	button.i_am_special = false;
	local value = button.title
	if (value == TRADES_LEATHERWORKING) then
		if (TradesBar_GetSpellID(TRADES_DRAGONSCALE_LEATHERWORKING) or TradesBar_GetSpellID(TRADES_ELEMENTAL_LEATHERWORKING) or TradesBar_GetSpellID(TRADES_TRIBAL_LEATHERWORKING)) then
			button.i_am_special = true;
		end
	elseif (value == TRADES_ENGINEERING) then
		if (TradesBar_GetSpellID(TRADES_GNOME_ENGINEERING) or  TradesBar_GetSpellID(TRADES_GOBLIN_ENGINEERING)) then
			button.i_am_special = true;
		end
	elseif (value == TRADES_TAILORING) then
		if (TradesBar_GetSpellID(TRADES_SHADOWEAVE_TAILORING) or TradesBar_GetSpellID(TRADES_SPELLFIRE_TAILORING)) then
			button.i_am_special = true;
		end
	end
end

function TradesBar_Update()
	if InCombatLockdown() then
		queueframe.queueitems.TradesBar_Update = true
		queueframe:Show()
		return
	end
	if (TradesBarData["settings"] and TradesBarData["settings"]["hidebar"] == "false") then
		local i = 0;
		for key in ipairs(Trades) do
			if TradesBar_Buttons[key] then
				TradesBar_Buttons[key]:Hide();
			end
		end
		TradesBar:Show();
		local height = 0
		for key in pairs(Trades) do
			local id = TradesBar_GetSpellID(Trades[key].title);
			TradesBar_SpecialCheck(Trades[key]);
			TradesBar_visableCheck(Trades[key]);
			if (not Trades[key].i_am_special and not Trades[key].hide_button) then
				if (id) then
--					ChatFrame1:AddMessage('Found '..Trades[key].title);
					i = i + 1;
					local button = TradesBar_Buttons[key];
					local texture = GetSpellTexture(id, BOOKTYPE_SPELL);
					_G[button:GetName().."Icon"]:SetTexture(texture);
					if button then
						button.id = id;
						button.texture = texture;
						if button.type == 'skill' then
							button:SetScript("OnClick",
								function(self, ...)
									CastSpellByName(self.title)
								end
							)
						end
						button:Show();
						height = height + 28
					end
				elseif Trades[key].type == 'item' then
--					ChatFrame1:AddMessage('Found '..Trades[key].title);
					i = i + 1;
					local button = TradesBar_Buttons[key];
					local _, _, _, _, _, _, _, _, _, texture = GetItemInfo(Trades[key].title);
					_G[button:GetName().."Icon"]:SetTexture(texture);
--					button.id = id;
					button.texture = texture;
					button:Show();
					height = height + 28
				elseif Trades[key].type == 'macro' then
					i = i + 1;
					local button = TradesBar_Buttons[key];
					_G[button:GetName().."Icon"]:SetTexture(Trades[key].texture);
					button.texture = Trades[key].texture;
					button.macrotext = Trades[key].macrotext;
					button.tooltiptext = Trades[key].tooltiptext;
					button:Show();
					height = height + 28
				elseif Trades[key].type == 'special' then
					local button = TradesBar_Buttons[key];
					if button.companions > 0 then
						i = i + 1;
						button:Show();
						height = height + 28
					end
				end
			else
--				button:Hide();
			end
		end
		if TradesBarData["settings"]["horizontal"] == 'true' then
			TradesBar:SetWidth(height)
			TradesBar:SetHeight(28)
		else
			TradesBar:SetWidth(28)
			TradesBar:SetHeight(height)
		end
		if (i == 0) then
			TradesBar:Hide();
		end
		TradesBar_O();
	end
end

function TradesBar_UpdateCooldowns()
	for i in ipairs(Trades) do
--		button = TradesBar_Buttons[i];
		if TradesBar_Buttons[i] then
			if (TradesBar_Buttons[i].id) then
				local cooldown = _G[TradesBar_Buttons[i]:GetName().."Cooldown"];
				local start, duration, enable = GetSpellCooldown(TradesBar_Buttons[i].id, 'spell');
				if (start > 0 and duration > 0) then
					CooldownFrame_SetTimer(cooldown, start, duration, enable);
				end
			elseif TradesBar_Buttons[i].type == 'item' and TradesBar_Buttons[i].itemLink then
				local cooldown = _G[TradesBar_Buttons[i]:GetName().."Cooldown"];
				local start, duration, enable = GetItemCooldown(TradesBar_Buttons[i].itemLink);
				if (start > 0 and duration > 0) then
					CooldownFrame_SetTimer(cooldown, start, duration, enable);
				end
			end
		end
	end
end

function TradesBar_GetSpellID(spell, rank)
	local i = 1;
	local spellID;
	local highestRank;
	while true do
		local spellName, spellRank = GetSpellName(i, "spell");
		if (not spellName) then
			break;
		end
		if (spellName == spell) then
			if (rank) then
				if (spellRank == rank) then
					if (TradesBar_Debug) then
						DEFAULT_CHAT_FRAME:AddMessage("Found |cffffffff'"..spell.." ("..rank..")'|r at slot |cffffffff'"..i.."'", 0.1, 1.0, 0.0);
					end
					return i;
				end
			else
				spellID = i;
				highestRank = spellRank;
			end
		end
		i = i + 1;
	end			
	if (TradesBar_Debug) then
		DEFAULT_CHAT_FRAME:AddMessage("Could not find |cffffffff'"..spell.." ("..rank..")'", 1.0, 0.1, 0.0);
	end
	return spellID, highestRank;
end

function TradesBar_GetSkillLineID(input)
	for i=1, GetNumSkillLines() do
		local skillName, _, _, skillRank, _, _, skillMaxRank = GetSkillLineInfo(i);
		if skillName == input then
			return i;
		end
	end
	-- if we've made it this far, something must have gone horribly wrong!
	return false;
end

function TradesBar_OnEnter(self)
	local skillColor, skillDiffPercent = '', 0;
	local x;
	x = self:GetRight();
	if ( x >= ( GetScreenWidth() / 2 ) ) then
		GameTooltip:SetOwner(self, "ANCHOR_LEFT");
	else
		GameTooltip:SetOwner(self, "ANCHOR_RIGHT");
	end
	if self.mountOrPet then
--		CompanionButton_OnEnter(self)
		GameTooltip:SetHyperlink("spell:"..self.spellID);
		GameTooltip:Show();
	elseif (self.id) then
		local skillName = GetSpellName(self.id, 'spell');
		tmpSkillName = TradesBar_ProperSkillName(skillName);
		if tmpSkillName ~= skillName then
			spellID = TradesBar_GetSpellID(tmpSkillName);
			skillName = tmpSkillName;
		else
			spellID = self.id;
		end
		if not self.spellID then
			self.spellID = TradesBar_GetSpellIDByName(skillName)
		end
		if self.spellID and GameTooltip.SetHyperlink then
			GameTooltip:SetHyperlink("spell:" .. self.spellID)
		else
			GameTooltip:SetSpell(self.id, 'spell');
		end
		local skillLineID, skillRank, skillMaxRank = TradesBar_GetSkillLineID(tmpSkillName), nil, nil;
		if skillLineID then
			_, _, _, skillRank, _, _, skillMaxRank = GetSkillLineInfo(skillLineID);
		end
--		GameTooltip_SetDefaultAnchor(GameTooltip, self);
		if (skillRank and skillMaxRank and skillName) then
			skillDiffPercent = (skillRank / skillMaxRank) * 100;
			skillColor = '';
			if (skillDiffPercent <= 25 ) then
				skillColor = 'ff0000';
			elseif (skillDiffPercent > 25 and skillDiffPercent <= 50) then
				skillColor = 'ff7f00';
			elseif (skillDiffPercent > 50 and skillDiffPercent <= 75) then
				skillColor = 'ffff00';
			elseif (skillDiffPercent > 75) then
				skillColor = '00ff00';
			end
			if not skillColor then
				skillColor = 'ff0000';
			end
--			DEFAULT_CHAT_FRAME:AddMessage('TradesBar: skillColor = "'..tostring(skillColor)..'", spellID = "'..tostring(spellID)..'", skillName = "'..tostring(skillName)..'", skillMaxRank = "'..tostring(skillMaxRank)..'", skillRank = "'..tostring(skillRank)..'"');
			TradesBar_TooltipRanksText[skillName] = {
				left = '|cffffffffSkill Level:|r';
				right = '|cff'..tostring(skillColor)..skillRank..'|r|cffffffff/'..skillMaxRank..'|r';
			};
			GameTooltip:AddLine(" ");
--			GameTooltip:AddDoubleLine(TradesBar_TooltipRanksText[skillName].left, TradesBar_TooltipRanksText[skillName].right);
			GameTooltip:AddDoubleLine('|cffffffffSkill Level:|r', '|cff'..skillColor..skillRank..'|r|cffffffff/'..skillMaxRank..'|r');
		end
		
		GameTooltip:Show();
	elseif self.type == 'item' and self.title then
		if not self.itemLink then
			_, self.itemLink = GetItemInfo(self.title)
		end
		if self.itemLink then
			GameTooltip:SetHyperlink(self.itemLink);
			GameTooltip:Show();
		end
	elseif self.type == 'macro' and self.tooltiptext then
		GameTooltip:SetText(self.tooltiptext)
		if TradesBar_Debug then
			local macro = SecureButton_GetModifiedAttribute(self, "macro", 'LeftButton');
			local macrotext = SecureButton_GetModifiedAttribute(self, "macrotext", 'LeftButton');
			GameTooltip:AddLine(tostring(macrotext).."\n"..tostring(macro));
		end
		GameTooltip:Show();
	end
end

function TradesBar_ProperSkillName(skillName)
	if (skillName == TRADES_DISENCHANT) then
		skillName = TRADES_ENCHANTING;
	elseif (skillName == TRADES_SMELTING or skillName == TRADES_FIND_MINERALS) then
		skillName = TRADES_MINING;
	elseif (skillName == TRADES_FIND_HERBS) then
		skillName = TRADES_HERBALISM;
	elseif (skillName == TRADES_BASIC_CAMPFIRE) then
		skillName = TRADES_COOKING;
	elseif (skillName == TRADES_PICK_LOCK or skillName == TRADES_PICK_POCKET) then
		skillName = TRADES_LOCKPICKING;
	elseif (skillName == TRADES_PROSPECTING) then
		skillName = TRADES_JEWELCRAFTING;
	elseif (skillName == TRADES_SHADOWEAVE_TAILORING or skillName == TRADES_SPELLFIRE_TAILORING) then
		skillName = TRADES_TAILORING
	elseif (skillName == TRADES_GNOME_ENGINEERING or skillName == TRADES_GOBLIN_ENGINEERING) then
		skillName = TRADES_ENGINEERING;
	elseif (skillName == TRADES_DRAGONSCALE_LEATHERWORKING or skillName == TRADES_ELEMENTAL_LEATHERWORKING or skillName == TRADES_TRIBAL_LEATHERWORKING) then
		skillName = TRADES_LEATHERWORKING;
	elseif (skillName == TRADES_MILLING) then
		skillName = TRADES_INSCRIPTION;
	end
	return skillName;
end

function TradesBar_OnBind(that_spell)
	CastSpellByName(that_spell);
end

function TradesBarGetPoint()
	local vLeft = TradesBar:GetLeft();
	local vTop = TradesBar:GetTop();
	TradesBarData["settings"]["left"] = vLeft;
	TradesBarData["settings"]["top"] = vTop;
--	DEFAULT_CHAT_FRAME:AddMessage(vLeft.." "..vTop);
end

-- CONFIG

function TradesBarConfig_OnHide(this)
	if original_heights['TradesBarConfigPanel4'] and original_heights['TradesBarConfigPanel5'] and original_heights['TradesBarConfig'] then
		TradesBarConfigPanel4:SetHeight(original_heights['TradesBarConfigPanel4']);
		TradesBarConfigPanel5:SetHeight(original_heights['TradesBarConfigPanel5']);
		TradesBarConfig:SetHeight(original_heights['TradesBarConfig']);
	end
--	self.offset = 0
end

local ret = {}
local count = function(t)
	table.wipe(ret)
	local index = 1
	for id, next in pairs(t) do
		ret[index] = 0
		for k, v in pairs(next) do
			ret[index] = ret[index] + 1
		end
		index = index + 1
	end
	return max(ret[1] or 0, ret[2] or 0, ret[3] or 0)
end

local toggle_page_buttons = function(offset)
	local prev, next = TradesBarConfigPrevPageButton, TradesBarConfigNextPageButton
	if offset < TRADESBAR_ADMIN_MAX_BUTTONS then
		prev:Disable()
	else
		prev:Enable()
	end
	if count(AdminPanels) - offset <= TRADESBAR_ADMIN_MAX_BUTTONS then
		next:Disable()
	else
		next:Enable()
	end
end

-- OnShow
function TradesBarConfig_OnShow(self)
	TradesBarConfig_OnHide(self)
	-- Top Panel
	local order, hidebar, horizontal;
	if (TradesBarData["settings"]["order"] == 'za') then
		order = true;
	else
		order = false;
	end
	if (TradesBarData["settings"]["hidebar"] == 'true') then
		hidebar = true;
	else
		hidebar = false;
	end
	if (TradesBarData["settings"]["horizontal"] == 'true') then
		horizontal = true;
	else
		horizontal = false;
	end
	TradesBarConfigPanel1CheckButton1:SetChecked( hidebar );
	TradesBarConfigPanel1CheckButton2:SetChecked( horizontal );
	TradesBarConfigPanel1CheckButton3:SetChecked( order );

	original_heights = {
		['TradesBarConfigPanel4'] = TradesBarConfigPanel4:GetHeight(),
		['TradesBarConfigPanel5'] = TradesBarConfigPanel5:GetHeight(),
		['TradesBarConfig'] = TradesBarConfig:GetHeight(),
	}

	TradesBarConfigPanel1SliderScale:SetValue( TradesBarData["settings"]["scale"] );
	
	local Point = {
		[4] = {
			x = 10,
			y = 0,
			height = TradesBarConfigPanel4:GetHeight(),
		},
		[5] = {
			x = 10,
			y = 0,
			height = TradesBarConfigPanel5:GetHeight(),
		},
	}
	
	local left, right = 1, 2;
	local color, r, g, b, a
	
	if not self.offset then
		self.offset = 0
	end
	toggle_page_buttons(self.offset)
	for k in pairs(AdminPanels) do
		if k ~= 6 and (k == 4 or k == 5) then
			local side = 1;
			for id, skillTable in pairs(AdminPanels[k]) do
				if AdminPanels[k][id].button then
					AdminPanels[k][id].button:Hide()
				end
				if k and id > self.offset and id <= self.offset + TRADESBAR_ADMIN_MAX_BUTTONS then
					local id = id - self.offset
					-- Alternate from left to right adding checkbuttons as we go,
					-- also increasing the size of the panel and its parent frame.
					if side == left then
						Point[k].x = 10;
						Point[k].y = Point[k].y - 27;
	--					Point[k].height = Point[k].height + 27; -- Increase the panel size
					elseif side == right then
						Point[k].x = 116;
					end
					local visible;
					if not skillTable.title then
						skillTable.title = UNKNOWN
					end
					if not TradesBarData["skills"][skillTable.title] then
						TradesBarData["skills"][skillTable.title] = {}
					end
					if (TradesBarData["skills"][skillTable.title].visible == 'true') then
						visible = true;
					else
						visible = false;
					end
					if AdminPanels[k][id] then
						if not _G['TradesBarConfigButton'..k..id] or not AdminPanels[k][id].button then
							AdminPanels[k][id].button = CreateFrame('CheckButton', 'TradesBarConfigButton'..k..id, _G['TradesBarConfigPanel'..k], 'TradesBarCheckButtonTemplate');
						end
						AdminPanels[k][id].button:ClearAllPoints()
						AdminPanels[k][id].button:SetPoint('TOPLEFT', _G['TradesBarConfigPanel'..k], Point[k].x, Point[k].y)
						AdminPanels[k][id].button:SetChecked(visible);
						-- Default is enabled, but if its disabled and you aquire the skill,
						-- the CheckButton will remain disabled until you relog unless you reenable it like self.
						if skillTable.subtype or ((skillTable.type and (skillTable.type == 'item' or skillTable.type == 'macro')) or TradesBar_HasSkill(skillTable.title)) then
							AdminPanels[k][id].button:Enable();
							color = "ffff00"
							r, g, b, a = 1, 1, 0, 1
						else
							AdminPanels[k][id].button:Disable();
							AdminPanels[k][id].button:SetChecked(false);
							color = "cccccc"
							r, g, b, a = .8, .8, .8, .3
						end
						local itemName
						if skillTable.type == 'item' then
							itemName = GetItemInfo(skillTable.title)
							AdminPanels[k][id].button.itemName = itemName
						else
							itemName = skillTable.title
						end
						_G['TradesBarConfigButton'..k..id..'Text']:SetText(--[["|cff" .. color ..]] skillTable.title --[[.. "|r"]]);
						_G['TradesBarConfigButton'..k..id..'Text']:SetTextColor(r, g, b, a)
						AdminPanels[k][id].button.title = skillTable.title
						AdminPanels[k][id].button:Show();
						side = 1 + mod(side, 2);
					end
				end
			end
		end
	end
--	TradesBarConfigPanel4:SetHeight(Point[4].height);
--	TradesBarConfigPanel5:SetHeight(Point[5].height);
	-- Add the current hard-coded size to the biggest panel height and make that the parent frame's height
--	TradesBarConfig:SetHeight(TradesBarConfig:GetHeight() + max(Point[5].height, Point[4].height));
end

function TradesBarConfigScale_OnClick(self)
	TradesBarData["settings"]["scale"] = self:GetValue();
	TradesBar_Scale();
end

function TradesBarConfigCheckBoxHide_OnClick()
	if (TradesBarData["settings"]["hidebar"] == "true") then
		TradesBarData["settings"]["hidebar"] = "false";
	else
		TradesBarData["settings"]["hidebar"] = "true";
	end
	TradesBar_Hide();
	TradesBar_Update();
end

function TradesBarConfigToggleAll_OnClick(ctype)
	for _, tab in pairs(Trades) do
		if tab.mountOrPet and tab.mountOrPet == ctype then
			if (TradesBarData["skills"][tab.title].visible == "true") then
				TradesBarData["skills"][tab.title].visible = "false";
			else
				TradesBarData["skills"][tab.title].visible = "true";
			end
		end
	end
	TradesBarData["settings"]["hide_all_"..ctype] = "false"
	TradesBar_Update();
	TradesBarConfig_OnShow(TradesBarConfig)
end

function TradesBarConfigDisableAll_OnClick(ctype)
	for _, tab in pairs(Trades) do
		if tab.mountOrPet and tab.mountOrPet == ctype then
			TradesBarData["skills"][tab.title].visible = "false";
		end
	end
	TradesBarData["settings"]["hide_all_"..ctype] = "true"
	TradesBar_Update();
	TradesBarConfig_OnShow(TradesBarConfig)
end

function TradesBarConfigCheckBox_OnClick()
	if (TradesBarData["settings"]["horizontal"] == "true") then
		TradesBarData["settings"]["horizontal"] = "false";
	else
		TradesBarData["settings"]["horizontal"] = "true"
	end
	TradesBar_O();
end

function TradesBarConfigCheckBoxOrder_OnClick()
	if (TradesBarData["settings"]["order"] == "az") then
		TradesBarData["settings"]["order"] = "za";
	else
		TradesBarData["settings"]["order"] = "az"
	end
	TradesBar_O();
end

function TradesBar_PrevPageButton_OnClick(self)
	local f = self:GetParent()
	if not f.offset then
		f.offset = 0
	else
		f.offset = f.offset - TRADESBAR_ADMIN_MAX_BUTTONS
	end
	if f.offset < 0 then
		f.offset = 0
	end
	f:Hide() -- seems kinda redundant, I know...
	f:Show()
end

function TradesBar_NextPageButton_OnClick(self)
	local f = self:GetParent()
	if not f.offset then
		f.offset = TRADESBAR_ADMIN_MAX_BUTTONS
	else
		f.offset = f.offset + TRADESBAR_ADMIN_MAX_BUTTONS
	end
	f:Hide() -- seems kinda redundant, I know...
	f:Show()
end

function TradesBarConfigCheckBox_ToggleSkill(self)
	local skill = self.title;
	local checked;
--	skill = 'TradesBarButton'..skill;
--	ChatFrame1:AddMessage('Toggling '..skill);
	if not TradesBarData["skills"][skill] then
		TradesBarData["skills"][skill] = {}
	end
	if (TradesBarData["skills"][skill].visible == "true") then
		TradesBarData["skills"][skill].visible = "false";
		checked = false;
	else
		TradesBarData["skills"][skill].visible = "true";
		checked = true;
	end
	if TradesBarData["skills"][skill].buttonType then
		TradesBarData["settings"]["hide_all_"..TradesBarData["skills"][skill].buttonType] = "false"
	end
	self:SetChecked(checked);
	if checked then
		_G[self:GetName().."Text"]:SetTextColor(1, 1, 0, 1)
	else
		_G[self:GetName().."Text"]:SetTextColor(.8, .8, .8, 1)
	end
	TradesBar_Update();
end

function TradesBarDump()
	TradesBarSkills_Table = AdminPanels
end

function TradesBar_GetSkills()
	for i = 1, 100000 do
		local spellName = GetSpellInfo(i);
		table.insert(TradesBarLocale, {
			['spellID'] = spellID, ['spellName'] = spellName,
		})
	end
end

function TradesBar_GetSpellIDByName(name)
	for i = 1, 100000 do
		local spellName = GetSpellInfo(i);
		if name == spellName then
			return i
		end
	end
end

